{ user interface for the MK85 emulator based on unfinished project written }
unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, IniFiles;

type
  TFormMK = class(TForm)
    FaceBox: TPaintBox;
    RefreshTimer: TTimer;
    CursorTimer: TTimer;
    RunTimer: TTimer;
    procedure RefreshTimerTimer(Sender: TObject);
    procedure FaceBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FaceBoxMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FaceBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure CursorTimerTimer(Sender: TObject);
    procedure RunTimerTimer(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormPaint(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
    FormMK: TFormMK;

implementation

{$R *.dfm}

uses
    Make_def, Cpu, Debug;

const
    Path = 'Skins\';
    FaceName = 'Face';
    CloakName = 'Cloak';
    OnName = 'On';
    OffName = 'Off';
    WhiteUpName = 'WhiteUp';
    WhiteDownName = 'WhiteDown';
    BlackUpName = 'BlackUp';
    BlackDownName = 'BlackDown';
    ExeUpName = 'ExeUp';
    ExeDownName = 'ExeDown';
    MK85Name = 'MK85';
    MName = 'M';

var
    BitMap, Face, MemFace: TBitMap;
    GraphMode: boolean = False;		{ key labels }

{ mouse }
    MD: Boolean;		{ True when the left mouse button was pressed
				  while it didn't point at any clickable
				  object (key, power switch). Intended action:
				  dragging the face image to another place. }
    XPos, YPos: Integer;	{ mouse coordinates when left button pressed }

{ LCD }
    BkVal: Byte;
    CurVis: Boolean = False;
    ViewMatrix: Boolean;	{ determines the LCD pixel size }
    Scr: array[0..95] of byte;	{ LCD shadow memory }

{ CPU }
    StartUpDelay: integer;
    CpuSpeed: integer;		{ how many instructions executes the emulated
				  CPU at each RunTimer call }


procedure DrawBmp (s: string; x, y: integer; t: boolean; What: Byte);
var
  filename: string;
  i, j: Integer;
  Btm: TBitMap;
begin
  filename := Concat (Path, s, '.bmp');
  if FileExists (filename) then
  begin
    Btm := TBitMap.Create;
    Btm.Transparent := t;
    BitMap.LoadFromFile (filename);
    Btm.Width := BitMap.Width;
    Btm.Height := BitMap.Height;
    Btm.Canvas.Draw (0, 0, BitMap);
    if What = 0 then
    begin
      for i := 0 to Btm.Width do
        for j := 0 to Btm.Height do
          if Btm.Canvas.Pixels[i,j] <> clWhite
            then begin
            Btm.Canvas.Pixels[i,j] := RGB (BkVal, BkVal, BkVal);
            end;
    end; {if}
    if t then Btm.TransparentColor := Btm.Canvas.Pixels [Btm.Width div 2, Btm.Height div 2];
    Face.Canvas.Draw (x, y, Btm);
  end
  else MessageDlg (Concat ('Failed to load the file ', filename), mtWarning, [mbOk], 0);
end {DrawBmp};


{ display a status sign }
procedure ViewHlp (
	s: string;	{ name of the bitmap file }
	x1: integer;
	y1: integer;	{ coordinates of the upper-left corner }
	what: byte);	{ draw image if not zero, clear if zero }
begin
  DrawBmp (s, x1, y1, False, What);
end {ViewHlp};


{ display a segment of the 7-segm. area }
procedure View7seg (
	x1: integer;
	y1: integer;	{ coordinates of the begin of the line }
	x2: integer;
	y2: integer;	{ coordinates of the end of the line }
	what: byte);	{ show segment if not zero, clear if zero }
begin
  if what <> 0 then Face.Canvas.Pen.Color := clBlack
               else Face.Canvas.Pen.Color := RGB (BkVal, BkVal, BkVal);
  Face.Canvas.MoveTo (x1, y1);
  Face.Canvas.LineTo (x2, y2);
end {View7seg};

procedure ClearScr;
var I: Integer;
begin
  for I := 0 to 95 do
    Scr[I] := not LCD[I];
end; {proc ClearScr}


{ draw the LCD contents }
procedure View;
var
  Pos, Row, Col, Index, SX, SY: Integer;
  B: byte;
begin

{ draw the matrix }
  for Pos := 0 to 11 do
  begin

    for Row := 1 to 7 do
    begin
      Index := Pos*8 + Row;
      B := lcd[Index];

{ handle the cursor }
      if CurVis and (Pos = (lcd[96] and $0F)) then
        if (Row = 7) or ((lcd[96] and $10) = 0) then B := $1F else B := 0;

      if (Scr[Index] <> B) then
      begin
        Scr[Index] := B;
        for Col := 0 to 4 do
        begin
          if (B and 1) <> 0 then Face.Canvas.Pen.Color := clBlack
                            else Face.Canvas.Pen.Color := RGB (BkVal, BkVal, BkVal);
          B := B shr 1;
          SX := 44 + Pos*14 + Col*2;
          if ViewMatrix then SY := 37 + Row*4
                        else SY := 43 + Row*3;
          Face.Canvas.MoveTo (SX, SY);
          if ViewMatrix then Face.Canvas.LineTo (SX, SY + 3)
                        else Face.Canvas.LineTo (SX, SY + 2);
        end {for};
      end {if};
    end {for};
  end {for};

{ draw the 7-segm. characters and the status signs }
  if Scr[$00] <> lcd[$00] then
  begin
    Scr[$00] := lcd[$00];
    ViewHlp ('EXT', 41, 29, lcd[$00] and $01);
    ViewHlp ('S', 60, 26, lcd[$00] and $02);
    ViewHlp ('F', 60, 32, lcd[$00] and $04);
  end {if};

  if Scr[$08] <> lcd[$08] then
  begin
    Scr[$08] := lcd[$08];
    ViewHlp ('RUN', 69, 26, lcd[$08] and $01);
    ViewHlp ('WRT', 69, 32, lcd[$08] and $02);
    ViewHlp ('DEG', 89, 29, lcd[$08] and $10);
  end {if};

  if Scr[$18] <> lcd[$18] then
  begin
    Scr[$18] := lcd[$18];
    ViewHlp ('RAD', 106, 29, lcd[$18] and $01);
  end {if};

  if Scr[$20] <> lcd[$20] then
  begin
    Scr[$20] := lcd[$20];
    ViewHlp ('GRA', 125, 29, lcd[$20] and $01);
  end {if};

  if Scr[$28] <> lcd[$28] then
  begin
    Scr[$28] := lcd[$28];
    ViewHlp ('TR', 144, 29, lcd[$28] and $01);
  end {if};

  if Scr[$30] <> lcd[$30] then
  begin
    Scr[$30] := lcd[$30];
    View7seg (159, 27, 159, 31, lcd[$30] and $01);
    View7seg (160, 31, 163, 31, lcd[$30] and $02);
    View7seg (159, 32, 159, 36, lcd[$30] and $04);
    View7seg (160, 36, 163, 36, lcd[$30] and $08);
    View7seg (163, 32, 163, 36, lcd[$30] and $10);
  end {if};

  if Scr[$38] <> lcd[$38] then
  begin
    Scr[$38] := lcd[$38];
    View7seg (163, 27, 163, 31, lcd[$38] and $01);
    View7seg (160, 26, 163, 26, lcd[$38] and $02);
    View7seg (166, 27, 166, 31, lcd[$38] and $04);
    View7seg (167, 31, 170, 31, lcd[$38] and $08);
    View7seg (166, 32, 166, 36, lcd[$38] and $10);
  end {if};

  if Scr[$40] <> lcd[$40] then
  begin
    Scr[$40] := lcd[$40];
    View7seg (167, 36, 170, 36, lcd[$40] and $01);
    View7seg (170, 32, 170, 36, lcd[$40] and $02);
    View7seg (170, 27, 170, 31, lcd[$40] and $04);
    View7seg (167, 26, 170, 26, lcd[$40] and $08);
    View7seg (173, 27, 173, 31, lcd[$40] and $10);
  end {if};

  if Scr[$48] <> lcd[$48] then
  begin
    Scr[$48] := lcd[$48];
    View7seg (174, 31, 177, 31, lcd[$48] and $01);
    View7seg (173, 32, 173, 36, lcd[$48] and $02);
    View7seg (174, 36, 177, 36, lcd[$48] and $04);
    View7seg (177, 32, 177, 36, lcd[$48] and $08);
    View7seg (177, 27, 177, 31, lcd[$48] and $10);
  end {if};

  if Scr[$50] <> lcd[$50] then
  begin
    Scr[$50] := lcd[$50];
    View7seg (174, 26, 177, 26, lcd[$50] and $01);
    View7seg (180, 27, 180, 31, lcd[$50] and $02);
    View7seg (181, 31, 184, 31, lcd[$50] and $04);
    View7seg (180, 32, 180, 36, lcd[$50] and $08);
    View7seg (181, 36, 184, 36, lcd[$50] and $10);
  end {if};

  if Scr[$58] <> lcd[$58] then
  begin
    Scr[$58] := lcd[$58];
    View7seg (184, 32, 184, 36, lcd[$58] and $01);
    View7seg (184, 27, 184, 31, lcd[$58] and $02);
    View7seg (181, 26, 184, 26, lcd[$58] and $04);
    ViewHlp ('STOP', 189, 29, lcd[$58] and $08);
  end {if};

end; {proc View}


procedure Init;
{ draw the initial calculator image }
var
  X: Integer;
begin
{ draw the background image }
  with Face do
  begin
    Width := 572;
    Height := 240;
    Canvas.Brush.Color := clSilver;
    Canvas.Pen.Color := clBlack;
    Canvas.Rectangle (0, 0, Width, Height);
  end {with};
  DrawBmp (FaceName, 0, 0, False, 1);
{ copy a selection of the face image to MemFace }
  with MemFace do
  begin
    Transparent := False;
    Width := 378;
    Height := 101;
    Canvas.Draw (-14, -131, Face);
  end {with};
{ select the calculator name MK85/MK85M depending on the RAM size }
  if RamSize <= MINRAMSIZE then X := 496 else
  begin
    X := 477;
    DrawBmp (MName, 543, 27, False, 1);
  end {if};
  DrawBmp (MK85Name, X, 27, False, 1);
{ clear the display memory }
  for X := 0 to 95 do
  begin
    Scr[X] := $1F;
    lcd[X] := $00;
  end {for};
end; {proc Init}


procedure TFormMK.RefreshTimerTimer(Sender: TObject);
begin
  View;
  FaceBox.Canvas.Draw (0, 0, Face);
end;


{ called when mouse button pressed }
procedure TFormMK.FaceBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
{ proceed only when left mouse button pressed }
  if Button = mbLeft then
  begin
    KeyCode1 := 0;

{ power switch }
    if (X >= 56) and (X <= 86) and (Y >= 106) and (Y <= 126) then
    begin
      DrawBmp (OffName, 56, 106, False, 1);
      KeyCode1 := 1;
    end

{ EXE key }
    else if (X >= 485) and (X <= 548) and (Y >= 203) and (Y <= 223) then
    begin
      DrawBmp (ExeDownName, 485, 203, True, 1);
      KeyCode1 := 58;
    end

{ white keys }
    else if (X >= 18) and (X <= 390) and (Y >= 107) and (Y <= 224) then
    begin
      if (((X - 18) mod 38) <= 30) and (((Y - 107) mod 33) <= 18) then
      begin
        KeyCode1 := (X - 18) div 38 + 10 * ((Y - 107) div 33);
        if KeyCode1 < 5 then KeyCode1 := 0	{ missing first 5 top keys }
        else DrawBmp (WhiteDownName,
		18 + 38 * (KeyCode1 mod 10),
		107 + 33 * (KeyCode1 div 10), True, 1);
      end {if};
    end

{ black keys }
    else if (X >= 411) and (X <= 548) and (Y >= 55) and (Y <= 223) then
    begin
      if (((X - 411) mod 37) <= 26) and (((Y - 55) mod 37) <= 20) then
      begin
        KeyCode1 := 40 + (X - 411) div 37 + 4 * ((Y - 55) div 37);
        if KeyCode1 > 57 then KeyCode1 := 0	{ missing last 2 bottom keys }
        else DrawBmp (BlackDownName,
		411 + 37 * ((KeyCode1 - 40) mod 4),
		55 + 37 * ((KeyCode1 - 40) div 4), True, 1);
      end {if};
    end {if};

    if KeyCode1 = 0 then
{ no valid key pressed, intended action: move the face image }
    begin
      XPos := X;
      YPos := Y;
      MD := True;
    end {if};

  end {if};
end {proc};


{ called when mouse button released }
procedure TFormMK.FaceBoxMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
{ proceed only when left mouse button pressed }
  if Button <> mbLeft then Exit;
  MD := False;
  case KeyCode1 of

{ power switch - close the application }
    1: if (X >= 56) and (X <= 86) and (Y >= 106) and (Y <= 126) then Close;

{ white keys }
    5..39: DrawBmp (WhiteUpName,
		18 + 38 * (KeyCode1 mod 10),
		107 + 33 * (KeyCode1 div 10), True, 1);

{ black keys }
    40..57: DrawBmp (BlackUpName,
		411 + 37 * ((KeyCode1 - 40) mod 4),
		55 + 37 * ((KeyCode1 - 40) div 4), True, 1);

{ EXE key }
    58: DrawBmp (ExeUpName, 485, 203, True, 1);
  end; {case}
  KeyCode1 := 0;
end;


{ called when moving the mouse while the button pressed }
procedure TFormMK.FaceBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var L, T: Integer;
    ST: Boolean;
begin
  if MD then	{ drag the face image }
  begin
    Left := Left + (X - XPos);
    Top := Top + (Y - YPos);
    Exit;
  end {if};

  ST := False;
  case KeyCode1 of

{ power switch }
    1: if (X < 56) or (X > 86) or (Y < 106) or (Y > 126) then
       DrawBmp (OnName, 56, 106, False, 1) else
       DrawBmp (OffName, 56, 106, False, 1);

{ white keys }
    5..39: begin {White Key}
             L := 18 + 38 * (KeyCode1 mod 10);
             T := 107 + 33 * (KeyCode1 div 10);
             if (X < L) or (X > L + 30) or (Y < T) or (Y > T + 18) then
             begin
               DrawBmp (WhiteUpName, L, T, True, 1);
               ST := True;
             end; {if}
           end; {5..39}

{ black keys }
    40..57: begin
              L := 411 + 37 * ((KeyCode1 - 40) mod 4);
              T := 55 + 37 * ((KeyCode1 - 40) div 4);
              if (X < L) or (X > L + 26) or (Y < T) or (Y > T + 20) then
              begin
                DrawBmp (BlackUpName, L, T, True, 1);
                ST := True;
              end; {if}
            end; {40..57}

{ EXE key }
    58: begin
          if (X < 485) or (X > 548) or (Y < 203) or (Y > 223) then
          begin
            DrawBmp (ExeUpName, 485, 203, True, 1);
            ST := True;
          end; {if}
        end; {58}

  end; {case}
  if ST then KeyCode1 := 0;
end;


procedure TFormMK.FormShow(Sender: TObject);
begin
  KeyCode1 := 0;
  KeyCode2 := 0;
  StartUpDelay := 5;
  CpuStop := False;
  CpuSteps := -1;
  BreakPoint := -1;
  Init;
  FaceBox.Width := Face.Width;
  FaceBox.Height := Face.Height;
  CpuReset;
  RefreshTimer.Enabled := True;
  CursorTimer.Enabled := True;
  RunTimer.Enabled := True;
end;


{ load the ROM and RAM images }
procedure MemLoad;
var
  f: file of byte;
  i: integer;
begin
{ load the ROM image }
  if FileExists ('ROM.BIN') then
  begin
    AssignFile (f, 'ROM.BIN');
    Reset (f);
    i := 0;
    while not Eof(f) and (i<ROMSIZE) do
    begin
      Read (f, rom[i]);
      i := i+1;
    end {while};
    CloseFile (f);
  end
  else MessageDlg ('Failed to load the file ROM.BIN', mtWarning, [mbOk], 0);
{ load the RAM image }
  if FileExists ('RAM.BIN') then
  begin
    AssignFile (f, 'RAM.BIN');
    Reset (f);
    i := 0;
    while not Eof(f) and (i<RamSize) do
    begin
      Read (f, ram[i]);
      i := i+1;
    end {while};
    CloseFile (f);
  end;
end {MemLoad};


procedure IniLoad;
var
  IniMK: TIniFile;
begin
  IniMK := TIniFile.Create (ExpandFileName('MK85M.INI'));
  with IniMK do
  begin
    ViewMatrix := ReadBool ('Settings', 'ViewMatrix', False);
    CpuSpeed := ReadInteger ('Settings', 'CpuSpeed', 100);
    RamSize := ReadInteger ('Settings', 'RamSize', 6144);
    BkVal := ReadInteger ('Settings', 'BkVal', 240);
  end {with};
  IniMK.Free;
end {IniLoad};

procedure IniSave;
var
  IniMK: TIniFile;
begin
  IniMK := TIniFile.Create (ExpandFileName('MK85M.INI'));
  with IniMK do
  begin
    WriteBool ('Settings', 'ViewMatrix', ViewMatrix);
    WriteInteger ('Settings', 'CpuSpeed', CpuSpeed);
//    WriteInteger ('Settings', 'RamSize', RamSize);
    WriteInteger ('Settings', 'BkVal', BkVal);
  end {with};
  IniMK.Free;
end {IniLoad};


{ initialise the application }
procedure TFormMK.FormCreate(Sender: TObject);
begin
  BitMap := TBitMap.Create;
  Face := TBitMap.Create;
  MemFace := TBitMap.Create;
  IniLoad;
  if RamSize < MINRAMSIZE then RamSize := MINRAMSIZE;
  if RamSize > MAXRAMSIZE then RamSize := MAXRAMSIZE;
  RamEnd := RAMSTART + RamSize;
  GetMem (ram, RamSize);
  MemLoad;
end;


{ save the RAM image to the file RAM.BIN }
procedure MemSave;
var
  f: file of byte;
  i: integer;
begin
  {$I-}
  AssignFile (f, 'RAM.BIN');
  Rewrite (f);
  for i := 0 to RamSize-1 do Write (f, ram[i]);
  CloseFile (f);
  {$I+}
  if IOResult <> 0 then MessageDlg ('Failed to save the file RAM.BIN',
    mtWarning, [mbOk], 0);
end {MemSave};


{ terminate the application }
procedure TFormMK.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  MemSave;
  IniSave;
  BitMap.Free;
  Face.Free;
  MemFace.Free;
  FreeMem (ram, RamSize);
end;


procedure TFormMK.CursorTimerTimer(Sender: TObject);
begin
  CurVis := not CurVis;
end;


procedure TFormMK.FormKeyPress(Sender: TObject; var Key: Char);
const
{ key codes 10 to 57 }
  Letters: string[48] = 'QWERTYUIOPASDFGHJKLaZXCVBNM =aaaa/789*456-123+0.';
var
  i: integer;
begin
  i := 1;
  Key := UpCase(Key);
  while (i <= 48) and (Key <> Letters[i]) do i := i+1;
  if i <= 48 then KeyCode2 := i+9;
end;


procedure TFormMK.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_INSERT:	KeyCode2 := 5;	{ MODE }
    VK_LEFT:	KeyCode2 := 6;	{ <- }
    VK_RIGHT:	KeyCode2 := 7;	{ -> }
    VK_HOME:	KeyCode2 := 8;	{ [S] }
    VK_END:	KeyCode2 := 9;	{ [F] }
    VK_ESCAPE:	KeyCode2 := 40;	{ AC }
    VK_DELETE:	KeyCode2 := 41;	{ DEL }
    VK_PAUSE:   KeyCode2 := 42; { Stop }
    VK_RETURN:  KeyCode2 := 58;	{ EXE }
    VK_F10:	DebugForm.Show;
    VK_F12:     KeyCode2 := 2; { Default Setup }

    VK_PRIOR:   if BkVal < 255 then
                begin
                  Inc (BkVal);
                  ClearScr;
                end; {if}
    VK_NEXT:    if BkVal > 0 then
                begin
                  Dec (BkVal);
                  ClearScr;
                end; {if}
    VK_F4:    begin
                ViewMatrix := not ViewMatrix;
                Face.Canvas.Pen.Color := clWhite;
                Face.Canvas.Brush.Color := clWhite;
                Face.Canvas.Rectangle (44, 37, 207, 68);
                ClearScr;
              end; {VK_F4}
    VK_F5:    if CpuSpeed < 4096 then Inc (CpuSpeed);
    VK_F6:    if CpuSpeed > 1 then Dec (CpuSpeed);
    VK_F7:    CpuSpeed := 100;
    VK_F8:    CpuSpeed := 200;
  end {case};
end;


procedure TFormMK.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_TAB then
  begin
    if GraphMode then Face.Canvas.Draw (14, 131, MemFace)
                 else DrawBmp (CloakName, 14, 131, True, 1);
    GraphMode := not GraphMode;
  end;
  KeyCode2 := 0;
end;


procedure TFormMK.FormPaint(Sender: TObject);
begin
  FaceBox.Canvas.Draw (0, 0, Face);
end;


{ execute a bunch of machine code instructions }
procedure TFormMK.RunTimerTimer(Sender: TObject);
var
  i: integer;
begin
{ Delayed start of the machine code execution "fixes" a problem of distorted
  initial 'READY P0' message (missing rows of 5 black pixels). It appears that
  sometimes bytes from the rom table containing $1F were read incorrectly as
  $00, while other values were read fine. I'm not able to trace this bug any
  further. }
  if not CpuStop then
  begin
    if StartUpDelay > 0 then
      StartUpDelay := StartUpDelay - 1
    else
    begin
      for i := 1 to CpuSpeed do
      begin
        CpuRun;
        if CpuSteps > 0 then
        begin
          CpuSteps := CpuSteps-1;
          if CpuSteps = 0 then
          begin
            DebugForm.Show;
            break;
          end {if};
        end {if};
        if (BreakPoint >= 0) and (BreakPoint = ptrw(@reg[R7])^) then
        begin
          DebugForm.Show;
          break;
        end {if};
      end {for};
    end {if};
  end {if};
end;


end.
