{ common stuff }

unit Make_def;

interface

  type
    ptrb = ^byte;		{ unsigned 8-bit }
    ptrw = ^word;		{ unsigned 16-bit }
    ptrshort = ^shortint;	{ signed 8-bit }
    ptrsmall = ^smallint;	{ unsigned 16-bit }

  const
    digits: string[22] = '0123456789ABCDEFabcdef';

{ indexes to the 'reg' table }
    R0		= 0*2;
    R1		= 1*2;
    R2		= 2*2;
    R3		= 3*2;
    R4		= 4*2;
    R5		= 5*2;
    R6		= 6*2;
    R7		= 7*2;

{ status bits }
    I_bit	= $100;	{ interrupt in service flag }
    T_bit	= $10;
    N_bit	= $08;
    Z_bit	= $04;
    V_bit	= $02;
    C_bit	= $01;
    NZ_bit	= $0C;	{ N+Z }
    VC_bit	= $03;	{ V+C }
    NZV_bit	= $0E;	{ N+Z+V }
    NZVC_bit	= $0F;	{ N+Z+V+C }

{ ROM address space }
    ROMSTART	= $0000;
    ROMSIZE	= $4000;
    ROMEND	= ROMSTART + ROMSIZE;
{ RAM address space }
    RAMSTART	= $8000;
    MINRAMSIZE	= $0800;
    MAXRAMSIZE	= $4000;
{ LCD memory address space }
    LCDSTART    = $0080;
    LCDSIZE     = $0080;
    LCDEND      = $0100;

    dummysrc: array[0..2] of byte	{ free adress space }
	= ($FF, $FF, $FF);

  var
    dummydst: array[0..2] of byte;	{ free address space }
    dummyptr: pointer;
    rom: array[0..ROMSIZE-1] of byte;
    ram: array of byte;
    RamSize: word;
    RamEnd: word;			{ = RAMSTART + RamSize }
    reg: array[0..15] of byte;
    lcd: array[0..LCDSIZE-1] of byte;	{ LCD controller port }
    kbdcols: word;              	{ keyboard input port $0100 }
    kbdrows: word;			{ keyboard output port $0102 }
    cpuctrl: word;			{ CPU control register $0104 }
    psw: word;
    code: word;
    src, dst: pointer;

    KeyCode1: integer;			{ from the mouse }
    KeyCode2: integer;			{ from the keyboard }
					{ 0:	no key pressed
					  1:	power switch
					  2:	INIT on the back side
					  5-39:	white keys
					  40-57:black keys
					  58:	EXE key		}

    CpuStop: boolean;                   {True stops the CPU}
    CpuSteps: integer;			{ignored when < 0}
    BreakPoint: integer;		{ignored when < 0}

{ tables convert KeyCode1 and KeyCode2 to the keyboard columns state }
const

{ keyboard row 1 }
  KeyTab1: array[0..58] of word = (
	$0000,					{ no key pressed }
	$0000,					{ power switch }
	$0018,					{ INIT }
	$0000, $0000,				{ missing keys }
	$0000, $0000, $0000, $0000, $0000,	{ MODE <- -> [S] [F] }
	$0000, $0000, $0000, $0000, $0000,	{ Q W E R T }
	$0000, $0000, $0000, $0000, $0000,	{ Y U I O P }
	$0000, $0000, $0000, $0000, $0000,	{ A S D F G }
	$0000, $0000, $0000, $0000, $0000,	{ H J K L ANS }
	$0000, $0000, $0000, $0000, $0000,	{ Z X C V B }
	$0000, $0000, $0000, $0000, $0000,	{ N M spc = EE }
	$000C, $0014, $0000, $0024,		{ AC DEL STOP / }
	$0028, $0030, $0044, $0048,		{ 7 8 9 * }
	$0050, $0060, $0084, $0088,		{ 4 5 6 - }
	$0090, $00A0, $00C0, $0104,		{ 1 2 3 + }
	$0108, $0110, $0120 );			{ 0 . EXE }

{ keyboard row 2 }
  KeyTab2: array[0..58] of word = (
	$0000,					{ no key pressed }
	$0000,					{ power switch }
	$0000,					{ INIT }
	$0000, $0000,				{ missing keys }
	$000C, $0014, $0018, $0024, $0028,	{ MODE <- -> [S] [F] }
	$0000, $0000, $0000, $0000, $0000,	{ Q W E R T }
	$0030, $0044, $0048, $0050, $0060,	{ Y U I O P }
	$0000, $0000, $0000, $0000, $0000,	{ A S D F G }
	$0084, $0088, $0090, $00A0, $00C0,	{ H J K L ANS }
	$0000, $0000, $0000, $0000, $0000,	{ Z X C V B }
	$0108, $0104, $0110, $0120, $0140,	{ N M spc = EE }
	$0000, $0000, $0000, $0000,		{ AC DEL STOP / }
	$0000, $0000, $0000, $0000,		{ 7 8 9 * }
	$0000, $0000, $0000, $0000,		{ 4 5 6 - }
	$0000, $0000, $0000, $0000,		{ 1 2 3 + }
	$0000, $0000, $0000 );			{ 0 . EXE }

{ keyboard row 3 }
  KeyTab3: array[0..58] of word = (
	$0000,					{ no key pressed }
	$0000,					{ power switch }
	$0000,					{ INIT }
	$0000, $0000,				{ missing keys }
	$0000, $0000, $0000, $0000, $0000,	{ MODE <- -> [S] [F] }
	$0044, $000C, $0018, $0090, $00A0,	{ Q W E R T }
	$0000, $0000, $0000, $0000, $0000,	{ Y U I O P }
	$0088, $0014, $0050, $0024, $0028,	{ A S D F G }
	$0000, $0000, $0000, $0000, $0000,	{ H J K L ANS }
	$0030, $0048, $0084, $00C0, $0060,	{ Z X C V B }
	$0000, $0000, $0000, $0000, $0000,	{ N M spc = EE }
	$0000, $0000, $0000, $0000,		{ AC DEL STOP / }
	$0000, $0000, $0000, $0000,		{ 7 8 9 * }
	$0000, $0000, $0000, $0000,		{ 4 5 6 - }
	$0000, $0000, $0000, $0000,		{ 1 2 3 + }
	$0000, $0000, $0000 );			{ 0 . EXE }

{ the STOP key is selected by row 1 }


  function HexWord (x: word; c: char) : string;
  function GetDigit (c: char) : integer;
  function GetValue (s: string; base: integer) : integer;


implementation


{ hex conversion, doesn't show leading zeroes when c<>'0' }
function HexWord (x: word; c: char) : string;
var
  d: char;
  s: string[3];
  i: integer;
  n: word;
begin
  s := '';
  n := 4096;
  for i := 1 to 3 do
  begin
    d := digits[(x div n) + 1];
    if (c='0') or (d<>'0') then
    begin
      s := s + d;
      c := '0';
    end;
    x := x mod n;
    n := n div 16;
  end {for};
  HexWord := s + digits[(x div n) + 1];
end {HexWord};

{ value of a hex digit }
function GetDigit (c: char) : integer;
var
  i: integer;
begin
  i := 1;
  while (i<=22) and (c <> digits[i]) do i:=i+1;
  if i>16 then GetDigit := i-7 else GetDigit := i-1;
end {GetDigit};


{ value of a string }
function GetValue (s: string; base: integer) : integer;
var
  x, y, i: integer;
begin
  x := 0;
  i := 1;
{ skip leading spaces }
  while i <= Length(s) do
  begin
    if s[i] <> ' ' then break;
    i := i+1;
  end {while};
  while i <= Length(s) do
  begin
    y := GetDigit(s[i]);
    if y >= base then break;
    x := x*base + y;
    i := i+1;
  end;
  GetValue := x;
end {GetValue};


end.
