{ PDP-11 disassembler, based on program pinst.c written by Martin Minow }

unit pdp11dis;

interface

  var
   loc_addr: word;	{ location counter }

  function FetchWord : word;
  function ScanNTAB (x: integer) : integer;
  function Mnemonic (i: integer; size: char) : string;
  function Arguments (i: integer; x: word) : string;


implementation

  uses make_def;

  const

    NONE = 0;	{ halt }
    RTS = 1;	{ rts reg }
    DOUBLE = 2;	{ mov src,dst }
    SWBYTE = 3;	{ sop[b] dst }
    SINGLE = 4;	{ sop dst }
    JSR = 5;	{ jsr reg,dst }
    MUL = 6;	{ mul src,reg }
    BR = 7;	{ br addr }
    SOB = 8;	{ sob reg,addr }
    SPL = 9;	{ spl n }
    MARK = 10;	{ mark n }
    TRAP = 11;	{ trap n }
    CODE = 12;	{ se[cvnz] }


  type

    tab = record
      mask: word;
      op: word;
      str: string[5];
      kind: byte;
    end;

  const

    NTAB = 77;

    mnem: array[1..NTAB] of tab = (
      (	mask:$0000;	op:$0000;	str:'halt';	kind:NONE	),
      (	mask:$0000;	op:$0001;	str:'wait';	kind:NONE	),
      (	mask:$0000;	op:$0002;	str:'rti';	kind:NONE	),
      (	mask:$0000;	op:$0003;	str:'bpt';	kind:NONE	),
      (	mask:$0000;	op:$0004;	str:'iot';	kind:NONE	),
      (	mask:$0000;	op:$0005;	str:'reset';	kind:NONE	),
      (	mask:$0000;	op:$0006;	str:'rtt';	kind:NONE	),
      (	mask:$0007;	op:$0000;	str:'illop';	kind:NONE	),
      (	mask:$003F;	op:$0040;	str:'jmp';	kind:SINGLE	),
      (	mask:$0007;	op:$0080;	str:'rts';	kind:RTS	),
      (	mask:$000F;	op:$0088;	str:'illop';	kind:NONE	),
      (	mask:$0007;	op:$0098;	str:'spl';	kind:SPL	),
      (	mask:$0000;	op:$00A0;	str:'nop';	kind:NONE	),
      (	mask:$0000;	op:$00A1;	str:'clc';	kind:NONE	),
      (	mask:$000F;	op:$00A0;	str:'clear';	kind:CODE	),
      (	mask:$0000;	op:$00B1;	str:'sec';	kind:NONE	),
      (	mask:$000F;	op:$00B0;	str:'set';	kind:CODE	),
      (	mask:$003F;	op:$00C0;	str:'swab';	kind:SINGLE	),
      (	mask:$00FF;	op:$0100;	str:'br';	kind:BR		),
      (	mask:$00FF;	op:$0200;	str:'bne';	kind:BR		),
      (	mask:$00FF;	op:$0300;	str:'beq';	kind:BR		),
      (	mask:$00FF;	op:$0400;	str:'bge';	kind:BR		),
      (	mask:$00FF;	op:$0500;	str:'blt';	kind:BR		),
      (	mask:$00FF;	op:$0600;	str:'bgt';	kind:BR		),
      (	mask:$00FF;	op:$0700;	str:'ble';	kind:BR		),
      (	mask:$01FF;	op:$0800;	str:'jsr';	kind:JSR	),
      (	mask:$803F;	op:$0A00;	str:'clr';	kind:SWBYTE	),
      (	mask:$803F;	op:$0A40;	str:'com';	kind:SWBYTE	),
      (	mask:$803F;	op:$0A80;	str:'inc';	kind:SWBYTE	),
      (	mask:$803F;	op:$0AC0;	str:'dec';	kind:SWBYTE	),
      (	mask:$803F;	op:$0B00;	str:'neg';	kind:SWBYTE	),
      (	mask:$803F;	op:$0B40;	str:'adc';	kind:SWBYTE	),
      (	mask:$803F;	op:$0B80;	str:'sbc';	kind:SWBYTE	),
      (	mask:$803F;	op:$0BC0;	str:'tst';	kind:SWBYTE	),
      (	mask:$803F;	op:$0C00;	str:'ror';	kind:SWBYTE	),
      (	mask:$803F;	op:$0C40;	str:'rol';	kind:SWBYTE	),
      (	mask:$803F;	op:$0C80;	str:'asr';	kind:SWBYTE	),
      (	mask:$803F;	op:$0CC0;	str:'asl';	kind:SWBYTE	),
      (	mask:$003F;	op:$0D00;	str:'mark';	kind:MARK	),
      (	mask:$003F;	op:$0D40;	str:'mfpi';	kind:SINGLE	),
      (	mask:$003F;	op:$0D80;	str:'mtpi';	kind:SINGLE	),
      (	mask:$003F;	op:$0DC0;	str:'sxt';	kind:SINGLE	),
      (	mask:$01FF;	op:$0E00;	str:'illop';	kind:NONE	),
      (	mask:$8FFF;	op:$1000;	str:'mov';	kind:DOUBLE	),
      (	mask:$8FFF;	op:$2000;	str:'cmp';	kind:DOUBLE	),
      (	mask:$8FFF;	op:$3000;	str:'bit';	kind:DOUBLE	),
      (	mask:$8FFF;	op:$4000;	str:'bic';	kind:DOUBLE	),
      (	mask:$8FFF;	op:$5000;	str:'bis';	kind:DOUBLE	),
      (	mask:$0FFF;	op:$6000;	str:'add';	kind:DOUBLE	),
      (	mask:$0FFF;	op:$E000;	str:'sub';	kind:DOUBLE	),
      (	mask:$01FF;	op:$7000;	str:'mul';	kind:MUL	),
      (	mask:$01FF;	op:$7200;	str:'div';	kind:MUL	),
      (	mask:$01FF;	op:$7400;	str:'ash';	kind:MUL	),
      (	mask:$01FF;	op:$7600;	str:'ashc';	kind:MUL	),
      (	mask:$01FF;	op:$7800;	str:'xor';	kind:JSR	),
      (	mask:$0007;	op:$7A00;	str:'fadd';	kind:RTS	),
      (	mask:$0007;	op:$7A08;	str:'fsub';	kind:RTS	),
      (	mask:$0007;	op:$7A10;	str:'fmul';	kind:RTS	),
      (	mask:$0007;	op:$7A18;	str:'fdiv';	kind:RTS	),
      (	mask:$01FF;	op:$7E00;	str:'sob';	kind:SOB	),
      (	mask:$07FF;	op:$7800;	str:'illop';	kind:NONE	),
      (	mask:$00FF;	op:$8000;	str:'bpl';	kind:BR		),
      (	mask:$00FF;	op:$8100;	str:'bmi';	kind:BR		),
      (	mask:$00FF;	op:$8200;	str:'bhi';	kind:BR		),
      (	mask:$00FF;	op:$8300;	str:'blos';	kind:BR		),
      (	mask:$00FF;	op:$8400;	str:'bvc';	kind:BR		),
      (	mask:$00FF;	op:$8500;	str:'bvs';	kind:BR		),
      (	mask:$00FF;	op:$8600;	str:'bhis';	kind:BR		),
      (	mask:$00FF;	op:$8700;	str:'blo';	kind:BR		),
      (	mask:$00FF;	op:$8800;	str:'emt';	kind:TRAP	),
      (	mask:$00FF;	op:$8900;	str:'trap';	kind:TRAP	),
      (	mask:$003F;	op:$8D00;	str:'illop';	kind:NONE	),
      (	mask:$003F;	op:$8D40;	str:'mfpd';	kind:SINGLE	),
      (	mask:$003F;	op:$8D80;	str:'mtpd';	kind:SINGLE	),
      (	mask:$03FF;	op:$8C00;	str:'illop';	kind:NONE	),
      (	mask:$0FFF;	op:$F000;	str:'45fpp';	kind:NONE	),
      (	mask:$FFFF;	op:$0000;	str:'illop';	kind:NONE	)
    );


{ fetch a word from the memory, advance the location counter }
function FetchWord : word;
begin
  if loc_addr < ROMEND then
    FetchWord := ptrw(@rom[loc_addr])^
  else if (loc_addr >= RAMSTART) and (loc_addr < RamEnd) then
    FetchWord := ptrw(@ram[loc_addr-RAMSTART])^
  else
    FetchWord := 0;
  loc_addr := loc_addr + 2;
end {FetchWord};


{ returns the name of the register r }
function preg (r: word) : string;
begin
  r := r and 7;
  case r of
    7: preg := 'pc';
    6: preg := 'sp';
    else preg := 'r' + digits[r+1];
  end;
end {preg};


{ returns the addressing mode }
function paddr (m: word) : string;
var
  r: word;
  s: string[1];
begin
  r := m and $07;
  if (m and $08) = 0 then s := '' else s := '@';

  case m and $38 of

    $00:
      paddr := preg(r);

    $08:
      paddr := '(' + preg(r) + ')';

    $10, $18:
      if r = $07 then
        paddr := s + '#' + HexWord(FetchWord, ' ')
      else
        paddr := s + '(' + preg(r) + ')+';

    $20, $28:
      paddr := s + '-(' + preg(r) + ')';

    $30, $38:
      if r = $07 then
        paddr := s + HexWord(loc_addr + FetchWord, '0')
      else
        paddr := s + HexWord(FetchWord, '0') + '(' + preg(r) + ')';

  end {case};
end {paddr};


{ returns the names of the CCR register bits }
function ccr (x: word) : string;
var
  s: string[4];
begin
  s := '';
  if (x and 1) <> 0 then s := s + 'c';
  if (x and 2) <> 0 then s := s + 'v';
  if (x and 4) <> 0 then s := s + 'z';
  if (x and 8) <> 0 then s := s + 'n';
  ccr := s;
end {ccr};


{ returns the index to the NTAB table for op-code x }
function ScanNTAB (x: integer) : integer;
var
  i, j: integer;
begin
  i := 0;
  for j := 1 to NTAB do
  begin
    if (x and not mnem[j].mask) = mnem[j].op then
    begin
      i := j;
      break;
    end {if};
  end {for};
  ScanNTAB := i;
end {ScanNTAB};


{ returns the mnemonic }
function Mnemonic (
	i: integer;	{ index to the NTAB table }
	size: char)	{ 'b' for byte-size, otherwise ' ' }
	: string;
begin
  Result := mnem[i].str;
  if (mnem[i].kind = SWBYTE) or (mnem[i].kind = DOUBLE) then
    Result := Result + size;
end {Mnemonic};


{ returns the arguments }
function Arguments (
	i: integer;	{ index to the NTAB table }
	x: word)	{ op-code }
	: string;
begin
  case mnem[i].kind of

    SPL:
      Arguments := digits[(x and $07) + 1];

    RTS:
      Arguments := preg(x);

    SINGLE, SWBYTE:
      Arguments := paddr(x);

    DOUBLE:
      Arguments := paddr(x div 64) + ',' + paddr(x);

    JSR:
      Arguments := preg (x div 64) + ',' + paddr(x);

    MUL:
      Arguments := paddr(x) + ',' + preg(x div 64);

    BR:
      begin
        x := x and $FF;
        if (x and $80) <> 0 then x := x or $FF00;
        Arguments := HexWord(loc_addr + 2*x, '0');
      end;

    SOB:
      Arguments := preg(x div 64) + ',' + HexWord(loc_addr - 2*(x and $3F), '0');

    MARK:
      Arguments := HexWord(x and $3F, ' ');

    TRAP:
      Arguments := HexWord(x and $FF, ' ');

    CODE:
      Arguments := ccr(x);

    else
      Arguments := '';

  end {case};
end {Arguments};


end.
