{ PDP-11 addressing modes,
  based on the PDP-11/03 emulator written by Ovsienko V.A. }

unit srcdst;


interface

  var loc: word;	{ address of the resource }

  function GetSrcB (mode: word) : pointer;
  function GetSrcW (mode: word) : pointer;
  function DstPtr : pointer;


implementation

  uses make_def;

  type Func1 = function : pointer;

  var
    auto: word;		{ = 1 for byte size operations,
			 or 2 for word size operations }


{ function returns the pointer to the 'read' type resource at address 'loc' }
  function SrcPtr : pointer;
  var
    k: integer;
  begin
    if (loc>=RAMSTART) and (loc<RAMEND) then
	SrcPtr := @ram[loc-RAMSTART]
    else if (loc<$0100) or ((loc>=$0108) and (loc<ROMEND)) then
	SrcPtr := @rom[loc]
    else if loc=$0100 then
    begin
      if KeyCode2 <> 0 then k := KeyCode2 else k := KeyCode1;
      kbdcols := 0;
      if (kbdrows and $0002) <> 0 then kbdcols := kbdcols or KeyTab1[k];
      if (kbdrows and $0004) <> 0 then kbdcols := kbdcols or KeyTab2[k];
      if (kbdrows and $0008) <> 0 then kbdcols := kbdcols or KeyTab3[k];
      SrcPtr := @kbdcols;
    end
    else if loc=$0104 then SrcPtr := @cpuctrl
    else SrcPtr := @dummysrc;
  end {SrcPtr};

{ function returns the pointer to the 'write' type resource at address 'loc' }
  function DstPtr : pointer;
  begin
    if (loc>=RAMSTART) and (loc<RamEnd) then
	DstPtr := @ram[loc-RAMSTART]
    else if (loc>=LCDSTART) and (loc<LCDEND) then
      DstPtr := @lcd[loc-LCDSTART]
    else if loc=$0102 then DstPtr := @kbdrows
    else if loc=$0104 then DstPtr := @cpuctrl
    else DstPtr := @dummydst;
  end {DstPtr};


{ functions calculate the src/dst pointer,
  return the address, or 0 if the src/dst points to a register }

  function AddrMode_000 : pointer;	{ R0 }
  begin
    AddrMode_000 := @reg[R0];
    loc := 0;
  end;

  function AddrMode_001 : pointer;	{ R1 }
  begin
    AddrMode_001 := @reg[R1];
    loc := 0;
  end;

  function AddrMode_002 : pointer;	{ R2 }
  begin
    AddrMode_002 := @reg[R2];
    loc := 0;
  end;

  function AddrMode_003 : pointer;	{ R3 }
  begin
    AddrMode_003 := @reg[R3];
    loc := 0;
  end;

  function AddrMode_004 : pointer;	{ R4 }
  begin
    AddrMode_004 := @reg[R4];
    loc := 0;
  end;

  function AddrMode_005 : pointer;	{ R5 }
  begin
    AddrMode_005 := @reg[R5];
    loc := 0;
  end;

  function AddrMode_006 : pointer;	{ R6 }
  begin
    AddrMode_006 := @reg[R6];
    loc := 0;
  end;

  function AddrMode_007 : pointer;	{ R7 }
  begin
    AddrMode_007 := @reg[R7];
    loc := 0;
  end;



  function AddrMode_010 : pointer;	{ @R0 }
  begin
    loc := ptrw(@reg[R0])^;
    AddrMode_010 := SrcPtr;
  end {AddrMode_010};

  function AddrMode_011 : pointer;	{ @R1 }
  begin
    loc := ptrw(@reg[R1])^;
    AddrMode_011 := SrcPtr;
  end {AddrMode_011};

  function AddrMode_012 : pointer;	{ @R2 }
  begin
    loc := ptrw(@reg[R2])^;
    AddrMode_012 := SrcPtr;
  end {AddrMode_012};

  function AddrMode_013 : pointer;	{ @R3 }
  begin
    loc := ptrw(@reg[R3])^;
    AddrMode_013 := SrcPtr;
  end {AddrMode_013};

  function AddrMode_014 : pointer;	{ @R4 }
  begin
    loc := ptrw(@reg[R4])^;
    AddrMode_014 := SrcPtr;
  end {AddrMode_014};

  function AddrMode_015 : pointer;	{ @R5 }
  begin
    loc := ptrw(@reg[R5])^;
    AddrMode_015 := SrcPtr;
  end {AddrMode_015};

  function AddrMode_016 : pointer;	{ @R6 }
  begin
    loc := ptrw(@reg[R6])^;
    AddrMode_016 := SrcPtr;
  end {AddrMode_016};

  function AddrMode_017 : pointer;	{ @R7 }
  begin
    loc := ptrw(@reg[R7])^;
    AddrMode_017 := SrcPtr;
  end {AddrMode_017};



  function AddrMode_020 : pointer;	{ (R0)+ }
  begin
    loc := ptrw(@reg[R0])^;
    ptrw(@reg[R0])^ := ptrw(@reg[R0])^ + auto;
    AddrMode_020 := SrcPtr;
  end {AddrMode_020};

  function AddrMode_021 : pointer;	{ (R1)+ }
  begin
    loc := ptrw(@reg[R1])^;
    ptrw(@reg[R1])^ := ptrw(@reg[R1])^ + auto;
    AddrMode_021 := SrcPtr;
  end {AddrMode_021};

  function AddrMode_022 : pointer;	{ (R2)+ }
  begin
    loc := ptrw(@reg[R2])^;
    ptrw(@reg[R2])^ := ptrw(@reg[R2])^ + auto;
    AddrMode_022 := SrcPtr;
  end {AddrMode_022};

  function AddrMode_023 : pointer;	{ (R3)+ }
  begin
    loc := ptrw(@reg[R3])^;
    ptrw(@reg[R3])^ := ptrw(@reg[R3])^ + auto;
    AddrMode_023 := SrcPtr;
  end {AddrMode_023};

  function AddrMode_024 : pointer;	{ (R4)+ }
  begin
    loc := ptrw(@reg[R4])^;
    ptrw(@reg[R4])^ := ptrw(@reg[R4])^ + auto;
    AddrMode_024 := SrcPtr;
  end {AddrMode_024};

  function AddrMode_025 : pointer;	{ (R5)+ }
  begin
    loc := ptrw(@reg[R5])^;
    ptrw(@reg[R5])^ := ptrw(@reg[R5])^ + auto;
    AddrMode_025 := SrcPtr;
  end {AddrMode_025};

  function AddrMode_026 : pointer;	{ (R6)+ = (SP)+ }
  begin
    loc := ptrw(@reg[R6])^;
    ptrw(@reg[R6])^ := ptrw(@reg[R6])^ + 2;
    AddrMode_026 := SrcPtr;
  end {AddrMode_026};

  function AddrMode_027 : pointer;	{ (R7)+ = #n }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    AddrMode_027 := SrcPtr;
  end {AddrMode_027};



  function AddrMode_030 : pointer;	{ @(R0)+ }
  begin
    loc := ptrw(@reg[R0])^;
    ptrw(@reg[R0])^ := ptrw(@reg[R0])^ + 2;
    loc := ptrw(SrcPtr)^;
    AddrMode_030 := SrcPtr;
  end {AddrMode_030};

  function AddrMode_031 : pointer;	{ @(R1)+ }
  begin
    loc := ptrw(@reg[R1])^;
    ptrw(@reg[R1])^ := ptrw(@reg[R1])^ + 2;
    loc := ptrw(SrcPtr)^;
    AddrMode_031 := SrcPtr;
  end {AddrMode_031};

  function AddrMode_032 : pointer;	{ @(R2)+ }
  begin
    loc := ptrw(@reg[R2])^;
    ptrw(@reg[R2])^ := ptrw(@reg[R2])^ + 2;
    loc := ptrw(SrcPtr)^;
    AddrMode_032 := SrcPtr;
  end {AddrMode_032};

  function AddrMode_033 : pointer;	{ @(R3)+ }
  begin
    loc := ptrw(@reg[R3])^;
    ptrw(@reg[R3])^ := ptrw(@reg[R3])^ + 2;
    loc := ptrw(SrcPtr)^;
    AddrMode_033 := SrcPtr;
  end {AddrMode_033};

  function AddrMode_034 : pointer;	{ @(R4)+ }
  begin
    loc := ptrw(@reg[R4])^;
    ptrw(@reg[R4])^ := ptrw(@reg[R4])^ + 2;
    loc := ptrw(SrcPtr)^;
    AddrMode_034 := SrcPtr;
  end {AddrMode_034};

  function AddrMode_035 : pointer;	{ @(R5)+ }
  begin
    loc := ptrw(@reg[R5])^;
    ptrw(@reg[R5])^ := ptrw(@reg[R5])^ + 2;
    loc := ptrw(SrcPtr)^;
    AddrMode_035 := SrcPtr;
  end {AddrMode_035};

  function AddrMode_036 : pointer;	{ @(R6)+ = @(SP)+ }
  begin
    loc := ptrw(@reg[R6])^;
    ptrw(@reg[R6])^ := ptrw(@reg[R6])^ + 2;
    loc := ptrw(SrcPtr)^;
    AddrMode_036 := SrcPtr;
  end {AddrMode_036};

  function AddrMode_037 : pointer;	{ @(R7)+ = @#n when source type }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(SrcPtr)^;
    AddrMode_037 := SrcPtr;
  end {AddrMode_037};



  function AddrMode_040 : pointer;	{ -(R0) }
  begin
    ptrw(@reg[R0])^ := ptrw(@reg[R0])^ - auto;
    loc := ptrw(@reg[R0])^;
    AddrMode_040 := SrcPtr;
  end {AddrMode_040};

  function AddrMode_041 : pointer;	{ -(R1) }
  begin
    ptrw(@reg[R1])^ := ptrw(@reg[R1])^ - auto;
    loc := ptrw(@reg[R1])^;
    AddrMode_041 := SrcPtr;
  end {AddrMode_041};

  function AddrMode_042 : pointer;	{ -(R2) }
  begin
    ptrw(@reg[R2])^ := ptrw(@reg[R2])^ - auto;
    loc := ptrw(@reg[R2])^;
    AddrMode_042 := SrcPtr;
  end {AddrMode_042};

  function AddrMode_043 : pointer;	{ -(R3) }
  begin
    ptrw(@reg[R3])^ := ptrw(@reg[R3])^ - auto;
    loc := ptrw(@reg[R3])^;
    AddrMode_043 := SrcPtr;
  end {AddrMode_043};

  function AddrMode_044 : pointer;	{ -(R4) }
  begin
    ptrw(@reg[R4])^ := ptrw(@reg[R4])^ - auto;
    loc := ptrw(@reg[R4])^;
    AddrMode_044 := SrcPtr;
  end {AddrMode_044};

  function AddrMode_045 : pointer;	{ -(R5) }
  begin
    ptrw(@reg[R5])^ := ptrw(@reg[R5])^ - auto;
    loc := ptrw(@reg[R5])^;
    AddrMode_045 := SrcPtr;
  end {AddrMode_045};

  function AddrMode_046 : pointer;	{ -(R6) = -(SP) }
  begin
    ptrw(@reg[R6])^ := ptrw(@reg[R6])^ - 2;
    loc := ptrw(@reg[R6])^;
    AddrMode_046 := SrcPtr;
  end {AddrMode_046};

  function AddrMode_047 : pointer;	{ -(R7) = kind of pointless }
  begin
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ - 2;
    loc := ptrw(@reg[R7])^;
    AddrMode_047 := SrcPtr;
  end {AddrMode_047};



  function AddrMode_050 : pointer;	{ @-(R0) }
  begin
    ptrw(@reg[R0])^ := ptrw(@reg[R0])^ - 2;
    loc := ptrw(@reg[R0])^;
    loc := ptrw(SrcPtr)^;
    AddrMode_050 := SrcPtr;
  end {AddrMode_050};

  function AddrMode_051 : pointer;	{ @-(R1) }
  begin
    ptrw(@reg[R1])^ := ptrw(@reg[R1])^ - 2;
    loc := ptrw(@reg[R1])^;
    loc := ptrw(SrcPtr)^;
    AddrMode_051 := SrcPtr;
  end {AddrMode_051};

  function AddrMode_052 : pointer;	{ @-(R2) }
  begin
    ptrw(@reg[R2])^ := ptrw(@reg[R2])^ - 2;
    loc := ptrw(@reg[R2])^;
    loc := ptrw(SrcPtr)^;
    AddrMode_052 := SrcPtr;
  end {AddrMode_052};

  function AddrMode_053 : pointer;	{ @-(R3) }
  begin
    ptrw(@reg[R3])^ := ptrw(@reg[R3])^ - 2;
    loc := ptrw(@reg[R3])^;
    loc := ptrw(SrcPtr)^;
    AddrMode_053 := SrcPtr;
  end {AddrMode_053};

  function AddrMode_054 : pointer;	{ @-(R4) }
  begin
    ptrw(@reg[R4])^ := ptrw(@reg[R4])^ - 2;
    loc := ptrw(@reg[R4])^;
    loc := ptrw(SrcPtr)^;
    AddrMode_054 := SrcPtr;
  end {AddrMode_054};

  function AddrMode_055 : pointer;	{ @-(R5) }
  begin
    ptrw(@reg[R5])^ := ptrw(@reg[R5])^ - 2;
    loc := ptrw(@reg[R5])^;
    loc := ptrw(SrcPtr)^;
    AddrMode_055 := SrcPtr;
  end {AddrMode_055};

  function AddrMode_056 : pointer;	{ @-(R6) = @-(SP) }
  begin
    ptrw(@reg[R6])^ := ptrw(@reg[R6])^ - 2;
    loc := ptrw(@reg[R6])^;
    loc := ptrw(SrcPtr)^;
    AddrMode_056 := SrcPtr;
  end {AddrMode_056};

  function AddrMode_057 : pointer;	{ @-(R7) = kind of pointless }
  begin
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ - 2;
    loc := ptrw(@reg[R7])^;
    loc := ptrw(SrcPtr)^;
    AddrMode_057 := SrcPtr;
  end {AddrMode_057};



  function AddrMode_060 : pointer;	{ n(R0) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R0])^ + ptrw(SrcPtr)^;
    AddrMode_060 := SrcPtr;
  end {AddrMode_060};

  function AddrMode_061 : pointer;	{ n(R1) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R1])^ + ptrw(SrcPtr)^;
    AddrMode_061 := SrcPtr;
  end {AddrMode_061};

  function AddrMode_062 : pointer;	{ n(R2) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R2])^ + ptrw(SrcPtr)^;
    AddrMode_062 := SrcPtr;
  end {AddrMode_062};

  function AddrMode_063 : pointer;	{ n(R3) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R3])^ + ptrw(SrcPtr)^;
    AddrMode_063 := SrcPtr;
  end {AddrMode_063};

  function AddrMode_064 : pointer;	{ n(R4) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R4])^ + ptrw(SrcPtr)^;
    AddrMode_064 := SrcPtr;
  end {AddrMode_064};

  function AddrMode_065 : pointer;	{ n(R5) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R5])^ + ptrw(SrcPtr)^;
    AddrMode_065 := SrcPtr;
  end {AddrMode_065};

  function AddrMode_066 : pointer;	{ n(R6) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R6])^ + ptrw(SrcPtr)^;
    AddrMode_066 := SrcPtr;
  end {AddrMode_066};

  function AddrMode_067 : pointer;	{ n(R7) = n , relative direct }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R7])^ + ptrw(SrcPtr)^;
    AddrMode_067 := SrcPtr;
  end {AddrMode_067};



  function AddrMode_070 : pointer;	{ @n(R0) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R0])^ + ptrw(SrcPtr)^;
    loc := ptrw(SrcPtr)^;
    AddrMode_070 := SrcPtr;
  end {AddrMode_070};

  function AddrMode_071 : pointer;	{ @n(R1) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R1])^ + ptrw(SrcPtr)^;
    loc := ptrw(SrcPtr)^;
    AddrMode_071 := SrcPtr;
  end {AddrMode_071};

  function AddrMode_072 : pointer;	{ @n(R2) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R2])^ + ptrw(SrcPtr)^;
    loc := ptrw(SrcPtr)^;
    AddrMode_072 := SrcPtr;
  end {AddrMode_072};

  function AddrMode_073 : pointer;	{ @n(R3) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R3])^ + ptrw(SrcPtr)^;
    loc := ptrw(SrcPtr)^;
    AddrMode_073 := SrcPtr;
  end {AddrMode_073};

  function AddrMode_074 : pointer;	{ @n(R4) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R4])^ + ptrw(SrcPtr)^;
    loc := ptrw(SrcPtr)^;
    AddrMode_074 := SrcPtr;
  end {AddrMode_074};

  function AddrMode_075 : pointer;	{ @n(R5) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R5])^ + ptrw(SrcPtr)^;
    loc := ptrw(SrcPtr)^;
    AddrMode_075 := SrcPtr;
  end {AddrMode_075};

  function AddrMode_076 : pointer;	{ @n(R6) }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R6])^ + ptrw(SrcPtr)^;
    loc := ptrw(SrcPtr)^;
    AddrMode_076 := SrcPtr;
  end {AddrMode_076};

  function AddrMode_077 : pointer;	{ @n(R7) = @n , relative indirect }
  begin
    loc := ptrw(@reg[R7])^;
    ptrw(@reg[R7])^ := ptrw(@reg[R7])^ + 2;
    loc := ptrw(@reg[R7])^ + ptrw(SrcPtr)^;
    loc := ptrw(SrcPtr)^;
    AddrMode_077 := SrcPtr;
  end {AddrMode_077};



  const Tabl_AddrMode: array[0..63] of pointer = (
	@AddrMode_000,	@AddrMode_001,	@AddrMode_002,	@AddrMode_003,
	@AddrMode_004,	@AddrMode_005,	@AddrMode_006,	@AddrMode_007,
	@AddrMode_010,	@AddrMode_011,	@AddrMode_012,	@AddrMode_013,
	@AddrMode_014,	@AddrMode_015,	@AddrMode_016,	@AddrMode_017,
	@AddrMode_020,	@AddrMode_021,	@AddrMode_022,	@AddrMode_023,
	@AddrMode_024,	@AddrMode_025,	@AddrMode_026,	@AddrMode_027,
	@AddrMode_030,	@AddrMode_031,	@AddrMode_032,	@AddrMode_033,
	@AddrMode_034,	@AddrMode_035,	@AddrMode_036,	@AddrMode_037,
	@AddrMode_040,	@AddrMode_041,	@AddrMode_042,	@AddrMode_043,
	@AddrMode_044,	@AddrMode_045,	@AddrMode_046,	@AddrMode_047,
	@AddrMode_050,	@AddrMode_051,	@AddrMode_052,	@AddrMode_053,
	@AddrMode_054,	@AddrMode_055,	@AddrMode_056,	@AddrMode_057,
	@AddrMode_060,	@AddrMode_061,	@AddrMode_062,	@AddrMode_063,
	@AddrMode_064,	@AddrMode_065,	@AddrMode_066,	@AddrMode_067,
	@AddrMode_070,	@AddrMode_071,	@AddrMode_072,	@AddrMode_073,
	@AddrMode_074,	@AddrMode_075,	@AddrMode_076,	@AddrMode_077
	);


  function GetSrcB (mode: word) : pointer;
  begin
    auto := 1;
    GetSrcB := Func1(Tabl_AddrMode[mode]);
  end {GetSrcB};


  function GetSrcW (mode: word) : pointer;
  begin
    auto := 2;
    GetSrcW := Func1(Tabl_AddrMode[mode]);
  end {GetSrcW};

end.
